@@ -32,3 +32,9 @@ int charset_exists(int charset)
 {
     return charset_find_spec(charset) != NULL;
 }
+
+int charset_is_single_byte(int charset)
+{
+    charset_spec const *spec = charset_find_spec(charset);
+    return spec && spec->read == read_sbcs;
+}
