@@ -4,8 +4,8 @@
 
 #include <stdio.h>
 
-static char *licencetext[] = {
-    "Halibut is copyright (c) 1999-2004 Simon Tatham and James Aylett.",
+static const char *const licencetext[] = {
+    "Halibut is copyright (c) 1999-2017 Simon Tatham.",
     "",
     "Permission is hereby granted, free of charge, to any person",
     "obtaining a copy of this software and associated documentation files",
@@ -26,11 +26,26 @@ static char *licencetext[] = {
     "ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN",
     "CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE",
     "SOFTWARE.",
+    "",
+    "Halibut contains font metrics derived from the \"Font Metrics for PDF",
+    "Core 14 Fonts\", which carry the following copyright notice and licence:",
+    "",
+    "  Copyright (c) 1985, 1987, 1989, 1990, 1991, 1992, 1993, 1997",
+    "  Adobe Systems Incorporated.  All Rights Reserved.",
+    "",
+    "  This file and the 14 PostScript(R) AFM files it accompanies may be",
+    "  used, copied, and distributed for any purpose and without charge,",
+    "  with or without modification, provided that all copyright notices",
+    "  are retained; that the AFM files are not distributed without this",
+    "  file; that all modifications to this file or any of the AFM files",
+    "  are prominently noted in the modified file(s); and that this",
+    "  paragraph is not modified. Adobe Systems has no responsibility or",
+    "  obligation to support the use of the AFM files.",
     NULL
 };
 
 void licence(void) {
-    char **p;
+    const char *const *p;
     for (p = licencetext; *p; p++)
 	puts(*p);
 }
